import json
import re
import time
import traceback
import requests
import turbopuffer as tpuf
import gzip
from turbopuffer.error import AuthenticationError, raise_api_error
from typing import Optional, List


def find_api_key(api_key: Optional[str] = None) -> str:
    if api_key is not None:
        return api_key
    elif tpuf.api_key is not None:
        return tpuf.api_key
    else:
        raise AuthenticationError("No turbopuffer API key was provided.\n"
                                  "Set the TURBOPUFFER_API_KEY environment variable, "
                                  "or pass `api_key=` when creating a Namespace.")

def clean_base_url(base_url: str) -> str:
    if base_url.endswith(('/v1', '/v1/', '/')):
        return re.sub('(/v1|/v1/|/)$', '', base_url)
    else:
        return base_url


class Backend:
    api_key: str
    base_url: str
    session: requests.Session

    def __init__(self, api_key: Optional[str] = None, base_url: Optional[str] = None, headers: Optional[dict] = None):
        if base_url is None:
            base_url = tpuf.api_base_url

        self.api_key = find_api_key(api_key)
        self.base_url = clean_base_url(base_url)
        self.headers = headers
        self.session = requests.Session()
        self.session.headers.update({
            'Authorization': f'Bearer {self.api_key}',
            'User-Agent': f'tpuf-python/{tpuf.VERSION} {requests.utils.default_headers()["User-Agent"]}',
        })

        if headers is not None:
            self.session.headers.update(headers)

    def __eq__(self, other):
        if isinstance(other, Backend):
            return self.api_key == other.api_key and self.base_url == other.base_url and self.headers == other.headers
        else:
            return False

    def make_api_request(self,
                         *args: List[str],
                         method: Optional[str] = None,
                         query: Optional[dict] = None,
                         payload: Optional[dict] = None) -> dict:
        start = time.monotonic()
        if method is None and payload is not None:
            method = 'POST'

        request = requests.Request(method or 'GET', self.base_url + '/'.join(args))

        if query is not None:
            request.params = query

        performance = dict()
        if payload is not None:
            payload_start = time.monotonic()
            if isinstance(payload, dict):
                json_payload = tpuf.dump_json_bytes(payload)
                performance['json_time'] = time.monotonic() - payload_start
            elif isinstance(payload, bytes):
                json_payload = payload
            else:
                raise ValueError(f'Unsupported POST payload type: {type(payload)}')

            gzip_start = time.monotonic()
            gzip_payload = gzip.compress(json_payload, compresslevel=1)
            performance['gzip_time'] = time.monotonic() - gzip_start
            if len(gzip_payload) > 0:
                performance['gzip_ratio'] = len(json_payload) / len(gzip_payload)

            request.headers.update({
                'Content-Type': 'application/json',
                'Content-Encoding': 'gzip',
            })
            request.data = gzip_payload

        prepared = self.session.prepare_request(request)

        retry_attempt = 0
        timeouts = (tpuf.connect_timeout, tpuf.read_timeout)
        while retry_attempt < tpuf.max_retries:
            request_start = time.monotonic()
            try:
                # print(f'Sending request:', prepared.path_url, prepared.headers)
                response = self.session.send(prepared, allow_redirects=False, timeout=timeouts)
                performance['request_time'] = time.monotonic() - request_start
                # print(f'Request time (HTTP {response.status_code}):', performance['request_time'])

                if response.status_code >= 500 or response.status_code == 408 or response.status_code == 429:
                    response.raise_for_status()

                server_timing_str = response.headers.get('Server-Timing', '')
                if len(server_timing_str) > 0:
                    match_cache = re.match(r'.*cache;hit_ratio=([\d\.]+);temperature=(\w+)', server_timing_str)
                    if match_cache:
                        try:
                            performance['cache_hit_ratio'] = float(match_cache.group(1))
                            performance['cache_temperature'] = match_cache.group(2)
                        except ValueError:
                            pass
                    match_exhaustive_search = re.match(r'.*exhaustive_search;count=([\d\.]+)', server_timing_str)
                    if match_exhaustive_search:
                        try:
                            performance['exhaustive_search_count'] = int(match_exhaustive_search.group(1))
                        except ValueError:
                            pass
                    match_processing = re.match(r'.*processing_time;dur=([\d\.]+)', server_timing_str)
                    if match_processing:
                        try:
                            # todo: update this to `processing_time`?
                            performance['server_time'] = float(match_processing.group(1)) / 1000.0
                        except ValueError:
                            pass
                    match_query_execution = re.match(r'.*query_execution_time;dur=([\d\.]+)', server_timing_str)
                    if match_query_execution:
                        try:
                            performance['query_execution_time'] = float(match_query_execution.group(1)) / 1000.0
                        except ValueError:
                            pass

                if method == 'HEAD':
                    return dict(response.__dict__, **{
                        'performance': performance,
                    })

                content_type = response.headers.get('Content-Type', 'text/plain')
                if content_type == 'application/json':
                    try:
                        content = response.json()
                    except json.JSONDecodeError as err:
                        raise_api_error(response.status_code, traceback.format_exception_only(err), response.text)

                    if response.ok:
                        performance['total_time'] = time.monotonic() - start
                        return dict(response.__dict__, **{
                            'content': content,
                            'performance': performance,
                        })
                    else:
                        raise_api_error(response.status_code, content.get('status', 'error'), content.get('error', ''))
                else:
                    raise_api_error(response.status_code, 'Server returned non-JSON response', response.text)
            except (requests.HTTPError, requests.ConnectionError, requests.Timeout) as err:
                retry_attempt += 1
                # print(traceback.format_exc())
                if retry_attempt < tpuf.max_retries:
                    # print(f'Retrying request in {2 ** retry_attempt}s...')
                    time.sleep(2 ** retry_attempt)  # exponential falloff up to 64 seconds for 6 retries.
                else:
                    print(f'Request failed after {retry_attempt} attempts...')

                    if isinstance(err, requests.HTTPError):
                        raise_api_error(err.response.status_code,
                                   f'Request to {err.request.url} failed after {retry_attempt} attempts',
                                   str(err))
                    else:
                        raise



import os
import sys
api_key = os.environ.get('TURBOPUFFER_API_KEY')
api_base_url = os.environ.get('TURBOPUFFER_API_BASE_URL', 'https://api.turbopuffer.com')
encode_vectors_as_base64 = os.environ.get('TURBOPUFFER_ENCODE_VECTORS_AS_BASE64', 'true').lower() == 'true'
upsert_batch_size = 10_000
max_retries = 6
connect_timeout = 10 # seconds
read_timeout = 180 # seconds

try:
    import orjson  # extras = ["fast"]
    def dump_json_bytes(obj): return orjson.dumps(obj, option=orjson.OPT_SERIALIZE_NUMPY)
except ImportError:
    import json

    class NumpyEncoder(json.JSONEncoder):
        def default(self, obj):
            if 'numpy' in sys.modules:
                if isinstance(obj, sys.modules['numpy'].integer):
                    return int(obj)
                elif isinstance(obj, sys.modules['numpy'].floating):
                    return float(obj)
                elif isinstance(obj, sys.modules['numpy'].ndarray):
                    return obj.tolist()
            if isinstance(obj, VectorRow) or isinstance(obj, VectorColumns):
                # sanity check; should not be used in the v2 write api
                raise ValueError('Cannot serialize VectorRow or VectorColumns to JSON')
            return json.JSONEncoder.default(self, obj)

    def dump_json_bytes(obj): return json.dumps(obj, cls=NumpyEncoder).encode()

try:
    import pybase64  # extras = ["fast"]
    def b64encode_as_string(bytes): return pybase64.b64encode_as_string(bytes)
    def b64decode(string): return pybase64.b64decode(string, validate=True)
except ImportError:
    import base64
    def b64encode_as_string(bytes): return base64.b64encode(bytes).decode('utf-8')
    def b64decode(string): return base64.b64decode(string, validate=True)

from turbopuffer.version import VERSION
from turbopuffer.namespace import Namespace, namespaces, AttributeSchema, FullTextSearchParams
from turbopuffer.vectors import VectorColumns, VectorRow, VectorResult
from turbopuffer.query import VectorQuery, Filters, RankInput
from turbopuffer.error import TurbopufferError, AuthenticationError, APIError, NotFoundError



class TurbopufferError(Exception):
    pass


class AuthenticationError(TurbopufferError):
    pass


class APIError(TurbopufferError):
    def __init__(self, status_code: int, status_name: str, message: str):
        self.status_code = status_code
        self.status_name = status_name
        super().__init__(f'{status_name} (HTTP {status_code}): {message}')

class NotFoundError(APIError):
    pass

def raise_api_error(status_code, status_name, message):
    if status_code == 404:
        raise NotFoundError(status_code, status_name, message)
    else:
        raise APIError(status_code, status_name, message)



import sys
import iso8601
import json
from datetime import datetime
from turbopuffer.error import APIError
from turbopuffer.vectors import Cursor, VectorResult, VectorColumns, VectorRow, batch_iter, b64encode_vector
from turbopuffer.backend import Backend
from turbopuffer.query import VectorQuery, Filters, RankInput, ConsistencyDict
from typing import Any, Dict, List, Literal, Optional, Iterable, Union, overload
import turbopuffer as tpuf

CmekDict = Dict[Literal['key_name'], str]
EncryptionDict = Dict[Literal['cmek'], CmekDict]

class FullTextSearchParams:
    """
    Used for configuring BM25 full-text indexing for a given attribute.
    """

    language: str
    stemming: bool
    remove_stopwords: bool
    case_sensitive: bool

    def __init__(self, language: str, stemming: bool, remove_stopwords: bool, case_sensitive: bool):
        self.language = language
        self.stemming = stemming
        self.remove_stopwords = remove_stopwords
        self.case_sensitive = case_sensitive

    def as_dict(self) -> dict:
        return {
            "language": self.language,
            "stemming": self.stemming,
            "remove_stopwords": self.remove_stopwords,
            "case_sensitive": self.case_sensitive,
        }

class AttributeSchema:
    """
    The schema for a particular attribute within a namespace.
    """

    type: str # one of: 'string', 'uint', '[]string', '[]uint'
    filterable: bool
    full_text_search: Optional[FullTextSearchParams] = None

    def __init__(self, type: str, filterable: bool, full_text_search: Optional[FullTextSearchParams] = None):
        self.type = type
        self.filterable = filterable
        self.full_text_search = full_text_search

    def as_dict(self) -> dict:
        result = {
            "type": self.type,
            "filterable": self.filterable,
        }
        if self.full_text_search:
            result["full_text_search"] = self.full_text_search.as_dict()
        return result

# Type alias for a namespace schema
NamespaceSchema = Dict[str, AttributeSchema]

def parse_namespace_schema(data: dict) -> NamespaceSchema:
    namespace_schema = {}
    for key, value in data.items():
        fts_params = value.get('full_text_search')
        fts_instance = None
        if fts_params:
            fts_instance = FullTextSearchParams(
                language=fts_params['language'],
                stemming=fts_params['stemming'],
                remove_stopwords=fts_params['remove_stopwords'],
                case_sensitive=fts_params['case_sensitive']
            )
        attribute_schema = AttributeSchema(
            type=value['type'],
            filterable=value['filterable'],
            full_text_search=fts_instance
        )
        namespace_schema[key] = attribute_schema
    return namespace_schema

class Namespace:
    """
    The Namespace type represents a set of vectors stored in turbopuffer.
    Within a namespace, vectors are uniquely referred to by their ID.
    All vectors in a namespace must have the same dimensions.
    """

    name: str
    backend: Backend

    metadata: Optional[dict] = None

    def __init__(self, name: str, api_key: Optional[str] = None, base_url: Optional[str] = None, headers: Optional[dict] = None):
        """
        Creates a new turbopuffer.Namespace object for querying the turbopuffer API.

        This function does not make any API calls on its own.

        Specifying an api_key here will override the global configuration for API calls to this namespace.

        Specifying a base_url here will override the global configuration api_base_url for this namespace.
        """
        self.name = name
        self.backend = Backend(api_key, base_url, headers)

    def __str__(self) -> str:
        return f'tpuf-namespace:{self.name}'

    def __eq__(self, other):
        if isinstance(other, Namespace):
            return self.name == other.name and self.backend == other.backend
        else:
            return False

    def refresh_metadata(self):
        response = self.backend.make_api_request('/v1/namespaces', self.name, method='HEAD')
        status_code = response.get('status_code')
        if status_code == 200:
            headers = response.get('headers', dict())
            dimensions = int(headers.get('x-turbopuffer-dimensions', '0'))
            approx_count = int(headers.get('x-turbopuffer-approx-num-vectors', '0'))
            self.metadata = {
                'exists': dimensions != 0,
                'dimensions': dimensions,
                'approx_count': approx_count,
                'created_at': iso8601.parse_date(headers.get('x-turbopuffer-created-at')),
            }
        elif status_code == 404:
            self.metadata = {
                'exists': False,
                'dimensions': 0,
                'approx_count': 0,
                'created_at': None,
            }
        else:
            raise APIError(response.status_code, 'Unexpected status code', response.get('content'))

    def exists(self) -> bool:
        """
        Returns True if the namespace exists, and False if the namespace is missing or empty.
        """
        # Always refresh the exists check since metadata from namespaces() might be delayed.
        self.refresh_metadata()
        return self.metadata['exists']

    def dimensions(self) -> int:
        """
        Returns the number of vector dimensions stored in this namespace.
        """
        if self.metadata is None or 'dimensions' not in self.metadata:
            self.refresh_metadata()
        return self.metadata.pop('dimensions', 0)

    def approx_count(self) -> int:
        """
        Returns the approximate number of vectors stored in this namespace.
        """
        if self.metadata is None or 'approx_count' not in self.metadata:
            self.refresh_metadata()
        return self.metadata.pop('approx_count', 0)

    def created_at(self) -> Optional[datetime]:
        """
        Returns the creation date of this namespace.
        """
        if self.metadata is None or 'created_at' not in self.metadata:
            self.refresh_metadata()
        return self.metadata.pop('created_at', None)

    def schema(self) -> NamespaceSchema:
        """
        Returns the current schema for the namespace.
        """
        response = self.backend.make_api_request('/v1/namespaces', self.name, 'schema', method='GET')
        return parse_namespace_schema(response["content"])

    def update_schema(self, schema_updates: NamespaceSchema):
        """
        Writes updates to the schema for a namespace.
        Returns the final schema after updates are done.

        See https://turbopuffer.com/docs/schema for specifics on allowed updates.
        """
        request_payload = json.dumps({key: value.as_dict() for key, value in schema_updates.items()}).encode()
        response = self.backend.make_api_request('/v1/namespaces', self.name, 'schema', method='POST', payload=request_payload)
        return parse_namespace_schema(response["content"])

    def copy_from_namespace(self, source_namespace: str):
        """
        Copies all documents from another namespace to this namespace.

        See: https://turbopuffer.com/docs/upsert#parameters `copy_from_namespace`
        for specifics on how this works.
        """
        payload = {
            "copy_from_namespace": source_namespace
        }
        response = self.backend.make_api_request('/v2/namespaces', self.name, payload=payload)
        assert response.get('content', dict()).get('status', '') == 'OK', f'Invalid copy_from_namespace() response: {response}'

    def write(self,
          upsert_columns: Optional[Dict[str, List] | VectorColumns] = None,
          patch_columns: Optional[Dict[str, List]] = None,
          upsert_rows: Optional[List[Dict[str, Any] | VectorRow]] = None,
          patch_rows: Optional[List[Dict[str, Any]]] = None,
          deletes: Optional[List[Union[int, str]]] = None,
          delete_by_filter: Optional[Filters] = None,
          distance_metric: Optional[Literal["cosine_distance", "euclidean_squared"]] = None,
          schema: Optional[Dict] = None,
          encryption: Optional[EncryptionDict] = None) -> None:
        """
        Create, update, or delete documents.
        """
        payload = {}

        if upsert_columns is not None:
            payload["upsert_columns"] = encode_columns("upsert_columns", upsert_columns)

        if patch_columns is not None:
            payload["patch_columns"] = encode_columns("patch_columns", patch_columns)

        if upsert_rows is not None:
            payload["upsert_rows"] = encode_rows("upsert_rows", upsert_rows)

        if patch_rows is not None:
            payload["patch_rows"] = encode_rows("patch_rows", patch_rows)

        if deletes is not None:
            if not isinstance(deletes, List):
                raise ValueError("deletes must be a List")
            payload["deletes"] = deletes

        if delete_by_filter is not None:
            payload["delete_by_filter"] = delete_by_filter

        if distance_metric is not None:
            payload["distance_metric"] = distance_metric

        if schema is not None:
            payload["schema"] = schema

        if encryption is not None:
            payload["encryption"] = encryption

        response = self.backend.make_api_request('/v2/namespaces', self.name, payload=payload)
        response_content = response.get('content', dict())
        assert response_content.get('status', '') == 'OK', f'Invalid write() response: {response}'
        self.metadata = None  # Invalidate cached metadata
        return response_content.get('rows_affected')

    @overload
    def query(self,
              vector: Optional[List[float]] = None,
              distance_metric: Optional[str] = None,
              top_k: int = 10,
              include_vectors: bool = False,
              include_attributes: Optional[Union[List[str], bool]] = None,
              filters: Optional[Filters] = None,
              rank_by: Optional[RankInput] = None,
              consistency: Optional[ConsistencyDict] = None
              ) -> VectorResult:
        ...

    @overload
    def query(self, query_data: VectorQuery) -> VectorResult:
        ...

    @overload
    def query(self, query_data: dict) -> VectorResult:
        ...

    def query(self,
              query_data=None,
              vector=None,
              distance_metric=None,
              top_k=None,
              include_vectors=None,
              include_attributes=None,
              filters=None,
              rank_by=None,
              consistency=None,
              vector_encoding_format=None) -> VectorResult:
        """
        Searches vectors matching the search query.

        See https://turbopuffer.com/docs/reference/query for query filter parameters.
        """

        if query_data is None:
            return self.query(VectorQuery(
                vector=vector,
                distance_metric=distance_metric,
                top_k=top_k,
                include_vectors=include_vectors,
                include_attributes=include_attributes,
                filters=filters,
                rank_by=rank_by,
                consistency=consistency,
                vector_encoding_format=vector_encoding_format,
            ))
        if not isinstance(query_data, VectorQuery):
            if isinstance(query_data, dict):
                query_data = VectorQuery.from_dict(query_data)
            else:
                raise ValueError(f'query() input type must be compatible with turbopuffer.VectorQuery: {type(query_data)}')

        # If enabled, request vectors as base64-encoded strings.
        if tpuf.encode_vectors_as_base64 and query_data.include_vectors and query_data.vector_encoding_format is None:
            query_data.vector_encoding_format = 'base64'

        response = self.backend.make_api_request('/v1/namespaces', self.name, 'query', payload=query_data.__dict__)
        result = VectorResult(response.get('content', dict()), namespace=self)
        result.performance = response.get('performance')
        return result

    def vectors(self, cursor: Optional[Cursor] = None) -> VectorResult:
        """
        This function exports the entire dataset at full precision.
        A VectorResult is returned that will lazily load batches of vectors if treated as an Iterator.

        If you want to look up vectors by ID, use the query function with an id filter.
        """

        response = self.backend.make_api_request('/v1/namespaces', self.name, query={'cursor': cursor})
        content = response.get('content', dict())
        next_cursor = content.pop('next_cursor', None)
        result = VectorResult(content, namespace=self, next_cursor=next_cursor)
        result.performance = response.get('performance')
        return result

    def delete_all_indexes(self) -> None:
        """
        Deletes all indexes in a namespace.
        """

        response = self.backend.make_api_request('/v1/namespaces', self.name, 'index', method='DELETE')
        assert response.get('content', dict()).get('status', '') == 'ok', f'Invalid delete_all_indexes() response: {response}'

    def delete_all(self) -> None:
        """
        Deletes all data as well as all indexes.
        """

        response = self.backend.make_api_request('/v1/namespaces', self.name, method='DELETE')
        assert response.get('content', dict()).get('status', '') == 'ok', f'Invalid delete_all() response: {response}'
        self.metadata = None  # Invalidate cached metadata

    def recall(self, num=20, top_k=10) -> float:
        """
        This function evaluates the recall performance of ANN queries in this namespace.

        When you call this function, it selects 'num' random vectors that were previously inserted.
        For each of these vectors, it performs an ANN index search as well as a ground truth exhaustive search.

        Recall is calculated as the ratio of matching vectors between the two search results.
        """

        response = self.backend.make_api_request('/v1/namespaces', self.name, '_debug', 'recall', query={'num': num, 'top_k': top_k})
        content = response.get('content', dict())
        assert 'avg_recall' in content, f'Invalid recall() response: {response}'
        return float(content.get('avg_recall'))


class NamespaceIterator:
    """
    The VectorResult type represents a set of vectors that are the result of a query.

    A VectorResult can be treated as either a lazy iterator or a list by the user.
    Reading the length of the result will internally buffer the full result.
    """

    backend: Backend
    namespaces: List[Namespace] = []
    index: int = -1
    offset: int = 0
    next_cursor: Optional[Cursor] = None

    def __init__(self, backend: Backend, initial_set: Union[List[Namespace], List[dict]] = [], next_cursor: Optional[Cursor] = None):
        self.backend = backend
        self.index = -1
        self.offset = 0
        self.next_cursor = next_cursor

        if len(initial_set):
            if isinstance(initial_set[0], Namespace):
                self.namespaces = initial_set
            else:
                self.namespaces = NamespaceIterator.load_namespaces(backend.api_key, initial_set)

    def load_namespaces(api_key: Optional[str], initial_set: List[dict]) -> List[Namespace]:
        output = []
        for input in initial_set:
            ns = tpuf.Namespace(input['id'], api_key=api_key)
            ns.metadata = {
                'exists': True,
            }
            output.append(ns)

        return output

    def __str__(self) -> str:
        str_list = [ns.name for ns in self.namespaces]
        if not self.next_cursor and self.offset == 0:
            return str(str_list)
        else:
            return ("NamespaceIterator("
                    f"offset={self.offset}, "
                    f"next_cursor='{self.next_cursor}', "
                    f"namespaces={str_list})")

    def __len__(self) -> int:
        assert self.offset == 0, "Can't call len(NamespaceIterator) after iterating"
        assert self.index == -1, "Can't call len(NamespaceIterator) after iterating"
        if not self.next_cursor:
            return len(self.namespaces)
        else:
            it = iter(self)
            self.namespaces = [next for next in it]
            self.offset = 0
            self.index = -1
            self.next_cursor = None
            return len(self.namespaces)

    def __getitem__(self, index) -> VectorRow:
        if index >= len(self.namespaces) and self.next_cursor:
            it = iter(self)
            self.namespaces = [next for next in it]
            self.offset = 0
            self.index = -1
            self.next_cursor = None
        return self.namespaces[index]

    def __iter__(self) -> 'NamespaceIterator':
        assert self.offset == 0, "Can't iterate over NamespaceIterator multiple times"
        return NamespaceIterator(self.backend, self.namespaces, self.next_cursor)

    def __next__(self):
        if self.index + 1 < len(self.namespaces):
            self.index += 1
            return self.namespaces[self.index]
        elif self.next_cursor is None:
            raise StopIteration
        else:
            response = self.backend.make_api_request(
                '/v1/namespaces',
                query={'cursor': self.next_cursor}
            )
            content = response.get('content', dict())
            self.offset += len(self.namespaces)
            self.index = -1
            self.next_cursor = content.pop('next_cursor', None)
            self.namespaces = NamespaceIterator.load_namespaces(self.backend.api_key, content.pop('namespaces', list()))
            return self.__next__()


def namespaces(api_key: Optional[str] = None, base_url: Optional[str] = None) -> Iterable[Namespace]:
    """
    Lists all turbopuffer namespaces for a given api_key.
    If no api_key is provided, the globally configured API key will be used.
    If no base_url is provided, the globally configured api_base_url will be used.
    """
    backend = Backend(api_key, base_url)
    response = backend.make_api_request('/v1/namespaces')
    content = response.get('content', dict())
    next_cursor = content.pop('next_cursor', None)
    return NamespaceIterator(backend, content.pop('namespaces', list()), next_cursor)

def encode_row(field: str, row: Dict[str, Any] | VectorRow) -> dict:
    if isinstance(row, VectorRow):
        return row.to_dict_for_write()
    elif isinstance(row, Dict):
        if "id" not in row:
            raise ValueError(f"each row in {field} must have an id")

        if tpuf.encode_vectors_as_base64:
            vector = row.get('vector')
            if isinstance(vector, list):
                row = row.copy()
                row['vector'] = b64encode_vector(vector)

        return row
    else:
        raise ValueError(f"{field} must be a dict or VectorRow")

def encode_rows(field: str, rows: List[Dict[str, Any] | VectorRow]) -> List[dict]:
    if not isinstance(rows, List):
        raise ValueError(f"{field} must be a list")

    return [encode_row(field, row) for row in rows]

def encode_columns(field: str, column_dict: Dict[str, List] | VectorColumns) -> dict:
    if isinstance(column_dict, VectorColumns):
        output = column_dict.to_dict_for_write()
    elif isinstance(column_dict, Dict):
        output = column_dict
        if "id" not in output:
            raise ValueError(f"{field} must have an id column")

        if tpuf.encode_vectors_as_base64:
            vectors = output.get('vector')
            if isinstance(vectors, list):
                output = output.copy()
                output['vector'] = [
                    b64encode_vector(vector) if isinstance(vector, list) else vector
                    for vector in vectors
                ]
    else:
        raise ValueError(f"{field} must be a dict or VectorColumns")

    return output



from dataclasses import dataclass
from functools import lru_cache
import struct
import sys
import turbopuffer as tpuf
from typing import Optional, Union, List, Iterable, Dict, overload
from itertools import islice


def batch_iter(iterable, n):
    it = iter(iterable)
    while True:
        batch = list(islice(it, n))
        if not batch:
            return
        yield batch


class Cursor(str):
    pass


@lru_cache(maxsize=64)
def _get_struct(vector_length: int) -> struct.Struct:
    return struct.Struct(f'<{vector_length}f')


def b64encode_vector(vector: List[float]) -> str:
    """
    Encodes a list of floats into a base64 string representation of the float
    array in a little-endian binary format.

    Args:
        vector (List[float]): The list of floats to encode.

    Returns:
        str: The base64 encoded string representation of the vector.
    """
    packed = _get_struct(len(vector)).pack(*vector)
    return tpuf.b64encode_as_string(packed)


def b64decode_vector(encoded: str) -> List[float]:
    """
    Decodes a base64 string representation of a float array in a little-endian
    binary format into a list of floats.

    Args:
        encoded (str): The base64 encoded string representation of the vector.

    Returns:
        List[float]: The decoded list of floats.
    """
    packed = tpuf.b64decode(encoded)
    return list(_get_struct(len(packed)//4).unpack(packed))


@dataclass
class VectorRow:
    """
    The VectorRow type represents a single vector ID, along with its vector values and attributes.

    If the VectorRow is a response from a query, it may also have a distance weighting.
    """

    id: Union[int, str]
    vector: Optional[List[float]] = None
    attributes: Optional[Dict[str, Optional[Union[str, int]]]] = None
    dist: Optional[float] = None

    def from_dict(source: dict) -> 'VectorRow':
        vector = source.get('vector')
        if vector is not None:
            if isinstance(vector, list):
                pass # already in float format
            elif isinstance(vector, str):
                vector = b64decode_vector(vector)
            else:
                raise ValueError(f'Unsupported vector type: {type(vector)}')
        return VectorRow(
            id=source.get('id'),
            vector=vector,
            attributes=source.get('attributes'),
            dist=source.get('dist'),
        )

    def __post_init__(self):
        if not isinstance(self.id, int) and not isinstance(self.id, str):
            raise ValueError('VectorRow.id must be an int or str, got:', type(self.id))
        if self.vector is not None:
            if 'numpy' in sys.modules and isinstance(self.vector, sys.modules['numpy'].ndarray):
                if self.vector.ndim != 1:
                    raise ValueError(f'VectorRow.vector must be a 1d array, got {self.vector.ndim} dimensions')
            elif not isinstance(self.vector, list) and not isinstance(self.vector, str):
                raise ValueError('VectorRow.vector must be a list or string, got:', type(self.vector))
        if self.attributes is not None and not isinstance(self.attributes, dict):
            raise ValueError('VectorRow.attributes must be a dict, got:', type(self.attributes))

    def __str__(self) -> str:
        fields = []
        if isinstance(self.id, int):
            fields.append(f'id={self.id}')
        else:
            fields.append(f"id='{self.id}'")
        fields.append(f'vector={self.vector}')
        if self.attributes:
            fields.append(f'attributes={self.attributes}')
        if self.dist is not None:
            fields.append(f'dist={self.dist}')
        return f"VectorRow({', '.join(fields)})"

    def to_dict_for_write(self) -> dict:
        o = { 'id': self.id }

        if self.vector is not None:
            if tpuf.encode_vectors_as_base64 and isinstance(self.vector, list):
                o['vector'] = b64encode_vector(self.vector)
            else:
                o['vector'] = self.vector

        if self.attributes is not None:
            o.update(self.attributes)

        return o


@dataclass
class VectorColumns:
    """
    The VectorColumns type represents a set of vectors stored in a column-oriented layout with their attributes.

    If the VectorColumns is a response from a query, it may also have a set of distance weightings for each vector.
    """

    ids: Union[List[int], List[str]]
    vectors: Optional[List[Optional[Union[List[float], str]]]] = None
    attributes: Optional[Dict[str, List[Optional[Union[str, int]]]]] = None

    distances: Optional[List[float]] = None

    def from_dict(source: dict) -> 'VectorColumns':
        return VectorColumns(
            ids=source.get('ids') or [],
            vectors=source.get('vectors') or None,
            attributes=source.get('attributes'),
            distances=source.get('distances'),
        )

    def __post_init__(self):
        if 'numpy' in sys.modules and isinstance(self.ids, sys.modules['numpy'].ndarray):
            if self.ids.ndim != 1:
                raise ValueError(f'VectorColumns.ids must be a 1d array, got {self.ids.ndim} dimensions')
        elif not isinstance(self.ids, list):
            raise ValueError('VectorColumns.ids must be a list, got:', type(self.ids))

        if self.vectors is not None:
            if 'numpy' in sys.modules and isinstance(self.vectors, sys.modules['numpy'].ndarray):
                if self.vectors.ndim != 2:
                    raise ValueError(f'VectorColumns.ids must be a 2d array, got {self.vectors.ndim} dimensions')
            elif not isinstance(self.vectors, list):
                raise ValueError('VectorColumns.vectors must be a list, got:', type(self.vectors))
            if len(self.ids) != len(self.vectors):
                raise ValueError('VectorColumns.ids and VectorColumns.vectors must be the same length')

        if self.attributes is not None:
            if not isinstance(self.attributes, dict):
                raise ValueError('VectorColumns.attributes must be a dict, got:', type(self.attributes))
            for key, values in self.attributes.items():
                if not isinstance(values, list):
                    raise ValueError(f'VectorColumns.attributes[{key}] must be a list, got:', type(values))
                if len(values) != len(self.ids):
                    raise ValueError(f'VectorColumns.attributes[{key}] must be the same length as VectorColumns.ids')
        if self.distances is not None and len(self.distances) != len(self.ids):
            raise ValueError('VectorColumns.distances must be the same length as VectorColumns.ids')

    def __str__(self) -> str:
        fields = [
            f'ids={self.ids}',
            f'vectors={self.vectors}',
        ]
        if self.attributes:
            fields.append(f'attributes={self.attributes}')
        if self.distances:
            fields.append(f'distances={self.distances}')
        return f"VectorColumns({', '.join(fields)})"

    def __len__(self) -> int:
        return len(self.ids)

    def __getitem__(self, index) -> VectorRow:
        # This functions as the main mechanism for converting Columns to Rows
        vector = self.vectors[index] if self.vectors is not None else None
        row = VectorRow(self.ids[index], vector, dist=(self.distances and self.distances[index]))
        if self.attributes:
            row.attributes = dict()
            for key, values in self.attributes.items():
                if values and len(values) > index:
                    if values[index] is not None:
                        row.attributes[key] = values[index]
        return row

    def to_dict_for_write(self) -> dict:
        o = { 'id': self.ids }

        if self.vectors is not None:
            if tpuf.encode_vectors_as_base64:
                o['vector'] = [
                    b64encode_vector(vector) if isinstance(vector, list) else vector
                    for vector in self.vectors
                ]
            else:
                o['vector'] = self.vectors

        if self.attributes is not None:
            o.update(self.attributes)

        return o


SET_DATA = Union[Iterable[VectorRow], VectorColumns]
DATA = Union[Iterable[dict], dict, VectorRow, SET_DATA]


class VectorResult:
    """
    The VectorResult type represents a set of vectors that are the result of a query.

    A VectorResult can be treated as either a lazy iterator or a list by the user.
    Reading the length of the result will internally buffer the full result.
    """

    namespace: Optional['Namespace'] = None
    data: Optional[SET_DATA] = None
    index: int = -1
    offset: int = 0
    next_cursor: Optional[Cursor] = None

    performance: Optional[dict] = None

    def __init__(self, initial_data: Optional[DATA] = None, namespace: Optional['Namespace'] = None, next_cursor: Optional[Cursor] = None):
        self.namespace = namespace
        self.index = -1
        self.offset = 0
        self.next_cursor = next_cursor

        self.data = VectorResult.load_data(initial_data)

    def load_data(initial_data: DATA) -> SET_DATA:
        if initial_data:
            if isinstance(initial_data, list):
                if isinstance(initial_data[0], dict):
                    return [VectorRow.from_dict(row) for row in initial_data]
                elif isinstance(initial_data[0], VectorRow):
                    return initial_data
                else:
                    raise ValueError(f'Unsupported list data type: {type(initial_data[0])}')
            elif isinstance(initial_data, dict):
                return VectorColumns.from_dict(initial_data)
            elif isinstance(initial_data, VectorColumns):
                return initial_data
            elif isinstance(initial_data, Iterable):
                raise ValueError('VectorResult from Iterable not yet supported.')
            else:
                raise ValueError(f'Unsupported data type: {type(initial_data)}')

    def __str__(self) -> str:
        if not self.next_cursor and self.offset == 0:
            return str(self.data)
        else:
            return ("VectorResult("
                    f"namespace='{self.namespace.name}', "
                    f"offset={self.offset}, "
                    f"next_cursor='{self.next_cursor}', "
                    f"data={self.data})")

    def __len__(self) -> int:
        assert self.offset == 0, "Can't call len(VectorResult) after iterating"
        assert self.index == -1, "Can't call len(VectorResult) after iterating"
        if not self.next_cursor:
            if self.data is not None:
                return len(self.data)
            return 0
        else:
            it = iter(self)
            self.data = [next for next in it]
            self.offset = 0
            self.index = -1
            self.next_cursor = None
            return len(self.data)

    def __getitem__(self, index) -> VectorRow:
        if index >= len(self.data) and self.next_cursor:
            it = iter(self)
            self.data = [next for next in it]
            self.offset = 0
            self.index = -1
            self.next_cursor = None
        return self.data[index]

    def __iter__(self) -> 'VectorResult':
        assert self.offset == 0, "Can't iterate over VectorResult multiple times"
        return VectorResult(self.data, self.namespace, self.next_cursor)

    def __next__(self):
        if self.data is not None and self.index + 1 < len(self.data):
            self.index += 1
            return self.data[self.index]
        elif self.next_cursor is None:
            raise StopIteration
        else:
            response = self.namespace.backend.make_api_request(
                '/v1/namespaces',
                self.namespace.name,
                query={'cursor': self.next_cursor}
            )
            content = response.get('content', dict())
            self.offset += len(self.data)
            self.index = -1
            self.next_cursor = content.pop('next_cursor', None)
            self.data = VectorResult.load_data(content)
            return self.__next__()



from dataclasses import dataclass
import sys
from typing import Optional, List, Tuple, Union, Dict
from typing import Iterable
from typing_extensions import Literal

# Refer to turbopuffer docs for valid operator names
FilterOperator = str

FilterValue = Union[str, int, List[str], List[int]]
FilterCondition = Tuple[str, FilterOperator, FilterValue]

LegacyFilterCondition = Tuple[FilterOperator, FilterValue]
LegacyFilterDict = Dict[str, List[LegacyFilterCondition]]

Filters = Union[Tuple[str, List["Filters"]], FilterCondition, LegacyFilterDict]

RankInputTextQuery = Union[
    Tuple[str, Literal['BM25'], str],
    Tuple[Literal['Sum'], Iterable['RankInputTextQuery']],
    Tuple[Literal['Product'], Tuple[float, Iterable['RankInputTextQuery']]],
    Tuple[Literal['Product'], Tuple[Iterable['RankInputTextQuery'], float]],
]

AttributeOrdering = Union[Literal['asc'], Literal['desc']]

RankInputOrderByAttribute = Tuple[str, AttributeOrdering]

# Uses tuples for nice typing, but also allows arrays for backwards compatibility
RankInput = Union[
    RankInputTextQuery,
    RankInputOrderByAttribute,
    List[Union[str, List[str]]],
]

ConsistencyDict = Dict[Literal['level'], Literal['strong', 'eventual']]

@dataclass
class VectorQuery:
    vector: Optional[List[float]] = None
    distance_metric: Optional[str] = None
    top_k: int = 10
    include_vectors: bool = False
    include_attributes: Optional[Union[List[str], bool]] = None
    filters: Optional[Filters] = None
    rank_by: Optional[RankInput] = None
    consistency: Optional[ConsistencyDict] = None
    vector_encoding_format: Optional[str] = None

    def from_dict(source: dict) -> 'VectorQuery':
        return VectorQuery(
            vector=source.get('vector'),
            distance_metric=source.get('distance_metric'),
            top_k=source.get('top_k'),
            include_vectors=source.get('include_vectors'),
            include_attributes=source.get('include_attributes'),
            filters=source.get('filters'),
            rank_by=source.get('rank_by'),
            consistency=source.get('consistency'),
            vector_encoding_format=source.get('vector_encoding_format'),
        )

    def __post_init__(self):
        if self.vector is not None:
            if 'numpy' in sys.modules and isinstance(self.vector, sys.modules['numpy'].ndarray):
                if self.vector.ndim != 1:
                    raise ValueError(f'VectorQuery.vector must be a 1d array, got {self.vector.ndim} dimensions')
            elif not isinstance(self.vector, list):
                raise ValueError('VectorQuery.vector must be a list, got:', type(self.vector))
        if self.include_attributes is not None:
            if not isinstance(self.include_attributes, list) and not isinstance(self.include_attributes, bool):
                raise ValueError('VectorQuery.include_attributes must be a list or bool, got:', type(self.include_attributes))
        if self.filters is not None:
            if not isinstance(self.filters, dict) and not isinstance(self.filters, list) and not isinstance(self.filters, tuple):
                raise ValueError('VectorQuery.filters must be a dict, tuple, or list, got:', type(self.filters))
        if self.rank_by is not None:
            if not isinstance(self.rank_by, list) and not isinstance(self.rank_by, tuple):
                raise ValueError('VectorQuery.rank_by must be a list or tuple, got:', type(self.rank_by))
            for item in self.rank_by:
                if not isinstance(item, str) and not isinstance(item, list) and not isinstance(item, tuple):
                    raise ValueError('VectorQuery.rank_by elements must be strings, tuples or lists, got:', type(item))
        if self.consistency is not None:
            if not isinstance(self.consistency, dict) or 'level' not in self.consistency:
                raise ValueError("VectorQuery.consistency must be a dict with a 'level' key")
            if self.consistency['level'] not in ('strong', 'eventual'):
                raise ValueError("VectorQuery.consistency level must be 'strong' or 'eventual', got:", self.consistency['level'])
